"""
Main for nbsplit

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from typing import List
from argparse import _ArgumentGroup, SUPPRESS
import copy
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.arguments import split_csv, join_csv
from nbtoolbelt.counting import REQUIRED_CELL_FIELDS
from nbtoolbelt.printing import print_dict

TOOL = 'split'


class SplitTool(Tool):
    """Split each notebook.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Split Jupyter notebooks into code cells and other cells."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Split notebook nb into markdown, code, and raw cells.

        .. note::

          **Modifies**: ``nb``

          **Side-effect**: writes new notebook files

        :param nb: notebook to split
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        nb_result = {}
        freq = {'total': len(nb.cells)}
        nb_result_path = {}

        # split the notebook into a part for each cell type
        for cell_type in args.split_cell_types:
            nb_result[cell_type] = copy.deepcopy(nb)  # notebook for cell type
            nb_result[cell_type].cells = [cell for cell in nb.cells if cell.cell_type == cell_type]
            freq[cell_type] = len(nb_result[cell_type].cells)

        # print statistics
        if not args.quiet:
            print_dict(freq, 'Split cell statistics')
        self._aggregate['outputs'][-1].update(freq)

        # set up destination paths
        for cell_type in args.split_cell_types:
            name = nb_path.stem + getattr(args, 'split_' + cell_type + '_result_name') + nb_path.suffix
            nb_result_path[cell_type] = nb_path.with_name(name)
            if args.debug and freq[cell_type]:
                print('Destination of {} cells:'.format(cell_type), nb_result_path[cell_type])

        # only write parts that have cells
        result = [(nb_result[cell_type], nb_result_path[cell_type])
                  for cell_type in args.split_cell_types
                  if freq[cell_type]]

        return result

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-t', '--split-cell-types', default=SUPPRESS, type=split_csv,
                           metavar='TYPES',
                           help='comma-separated list of cell types to split'
                                " (default: '{}')".format(join_csv(self._args.split_cell_types)))

    def check_and_adjust_arguments(self) -> None:
        bad = set(self._args.split_cell_types).difference(REQUIRED_CELL_FIELDS.keys())
        if bad:
            raise ValueError('Invalid cell types with -t: {}'.format(join_csv(bad)))

    def print_tool_args(self) -> None:
        print('  Split cell types: {}'.format(join_csv(self._args.split_cell_types)))


def main(cli_args: List[str] = None):
    return SplitTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Configuration file handling

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Configuration files use JSON to make it easier to move the Jupyter configuration file mechanism in the future.

There is a top-level entry for global options with the package name as key.
Each tool has a tool-specific entry with key 'nb' + tool name; e.g. 'nbrun'.

The top-level configuration is loaded first; typically it has options that apply to multiple tools.
Then the tool-specific configuration is loaded on top of that (using dict update).

TODO: First the embedded configuration file is loaded (these are defaults).

TODO: Then a configuration file in the home directory is loaded, if present.

Finally configuration files mentioned on the command line are loaded.
These are processed during argument parsing.
"""

import sys
from ._version import package_name
from pathlib import Path
from argparse import Namespace
from typing import Any, Dict, Union
import json
from .printing import print_namespace

TEST = False

CONFIG_FILE_NAME = package_name + '.json'


def config_path() -> Path:
    """Return path to embedded config file.

    :return: path to embedded config file
    """
    return Path(__file__).resolve().parent / 'data' / CONFIG_FILE_NAME


def load_config_file(tool: str=package_name,
                     ns: Namespace=None,
                     file_path: Union[str, Path]=config_path(),
                     verbose: bool=False) -> Namespace:
    """Load configuration for tool from file and return update namespace.
    Silently ignores file if it does not exist.

    .. note:: **Modifies**: ``ns``

    :param tool: tool for which to load a configuration file
    :param ns: namespace to load configuration into
    :param file_path: file path to load configuration from
    :param verbose: whether to print some diagnostic info
    :return: updated ns
    """
    if ns is None:
        ns = Namespace()  # create a fresh one
        # NOTE: in an earlier version, ns had default parameter value Namespace()
        # But that default is created once, and shared by all subsequent tools.
        # And that hurts during testing.

    if type(file_path) is str:
        file_path = Path(file_path)  # TODO can this raise an exception?

    if file_path.exists():
        try:
            config_all = json.load(file_path.open(encoding='utf-8'))  # type: Dict[str, Dict[str, Any]]
        except Exception as e:
            print('Could not load configuration for {} in: {}'.format(tool, file_path), file=sys.stderr)
            print('{}: {}'.format(type(e).__name__, e), file=sys.stderr)
            sys.exit(1)  # TODO do this more graciously (but it may have been called from tool constructor)
        if verbose:
            print('Loading configuration for {} from: {}'.format(tool, file_path))
        if TEST:
            print('load_config_file(tool={}, fp={})'.format(tool, file_path))
            print('[load_config_file] config_all=\n{}'.format(json.dumps(config_all, indent=2)))
        if package_name in config_all:
            # copy the loaded configuration to the namespace
            for key, value in config_all[package_name].items():
                setattr(ns, key, value)  # overwrite or create the key-value pair
        if tool in config_all:
            # copy the loaded configuration to the namespace; unnecessary when tool == package_name
            for key, value in config_all[tool].items():
                setattr(ns, key, value)  # overwrite or create the key-value pair
        if TEST:
            print_namespace(ns, '[load_config_file] config for ' + tool)
    else:
        if verbose:
            print('Skipping configuration in: {} (not present)'.format(file_path))

    return ns


def load_config(tool: str=package_name, verbose: bool=False) -> Namespace:
    """Load configuration for tool.

    Successively load configuration files (if present) in:

    * the package itself (see src/nbtoolbelt/data/CONFIG_FILE_NAME)
    * /etc/CONFIG_FILE_NAME
    * ~/.CONFIG_FILE_NAME (a dot file in the home directory)

    .. note:: **Modifies**: ``ns``

    :param tool: tool for which to load a configuration file (default: top-level)
    :param verbose: whether to print some diagnostic info
    :return: updated ns
    """
    ns = load_config_file(tool, verbose=verbose)

    fp_etc = Path('/etc') / CONFIG_FILE_NAME
    load_config_file(tool, ns, fp_etc, verbose)

    fp_home = Path.home() / ('.' + CONFIG_FILE_NAME)
    load_config_file(tool, ns, fp_home, verbose)

    if verbose:
        print_namespace(ns, 'Configuration accumulated for ' + tool)

    return ns



"""
Main for nbclean

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup, SUPPRESS
from typing import List, Set
from nbtoolbelt.arguments import split_csv, NegatableAction, join_csv
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.counting import REQUIRED_NB_METADATA_FIELDS
from nbtoolbelt.cleaning import clean_nb
from nbtoolbelt.printing import print_dict

TOOL = 'clean'


class CleanTool(Tool):
    """Clean each notebook.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Clean Jupyter notebooks by deleting or resetting selected elements."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Clean notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to clean
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        # do the cleaning, and collect some statistics
        freq = clean_nb(nb, args)

        # show statistics
        if not args.quiet:
            print_dict(freq, 'Cleaning counts')
        self._aggregate['outputs'][-1].update(freq)

        # set up destination path
        if args.inplace:
            nb_clean_path = nb_path
        else:
            nb_clean_path = nb_path.with_name(nb_path.stem + args.clean_result_name + nb_path.suffix)

        if args.debug:
            print('Clean destination:', nb_clean_path)

        return [(nb, nb_clean_path)]

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-g', '--global',
                           dest='clean_notebook_metadata_fields', default=SUPPRESS, type=split_csv,
                           metavar='FIELDS',
                           help='comma-separated list of fields to remove from metadata of notebook' +
                                " (default: '{}')".format(join_csv(self._args.clean_notebook_metadata_fields)))
        group.add_argument('-m', '--metadata',
                           dest='clean_cell_metadata_fields', default=SUPPRESS, type=split_csv,
                           metavar='FIELDS',
                           help='comma-separated list of fields to remove from metadata of all cells' +
                                " (default: '{}')".format(join_csv(self._args.clean_cell_metadata_fields)))
        group.add_argument('-t', '--tags', dest='clean_tags', default=SUPPRESS, type=split_csv,
                           metavar='TAGS',
                           help='comma-separated list of tags to remove from all cells;'
                                " use '-m tags' to remove all tags" +
                                " (default: '{}')".format(join_csv(self._args.clean_tags)))
        group.add_argument('-e', '--empty-cells',
                           action=NegatableAction, dest='clean_empty_cells', default=SUPPRESS,
                           help='delete cells with empty source, i.e. with whitespace only' +
                                ' (default: {})'.format(self._args.clean_empty_cells))
        group.add_argument('-o', '--outputs',
                           action=NegatableAction, dest='clean_outputs', default=SUPPRESS,
                           help='clean all outputs from all code cells ' +
                                ' (default: {})'.format(self._args.clean_outputs))

    def check_and_adjust_arguments(self):
        args = self._args

        nmf = set(args.clean_notebook_metadata_fields)  # type: Set[str]
        if nmf:
            intersection = nmf.intersection(REQUIRED_NB_METADATA_FIELDS)
            if intersection:
                print('WARNING: {} are required notebook metadata fields'.format(intersection))
                print('  These will not be cleaned')
                nmf.difference_update(REQUIRED_NB_METADATA_FIELDS)

    def print_tool_args(self) -> None:
        args = self._args

        if args.clean_notebook_metadata_fields:
            print("  Deleting notebook metadata fields: {}".format(args.clean_notebook_metadata_fields))
        if args.clean_cell_metadata_fields:
            print("  Deleting cell metadata fields: {}".format(args.clean_cell_metadata_fields))
        if args.clean_tags:
            print("  Deleting tags: {}".format(args.clean_tags))
        if args.clean_empty_cells:
            print('  Deleting cells with empty source')
        if args.clean_outputs:
            print('  Deleting outputs from all code cells')


def main(cli_args: List[str] = None):
    return CleanTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Tool nbdump; dump cell sources of notebooks

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup, SUPPRESS
from typing import List
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.arguments import NegatableAction, split_csv, join_csv
from nbtoolbelt.counting import REQUIRED_CELL_FIELDS, nb_other_metadata

TOOL = 'dump'


class DumpTool(Tool):
    """Show all source lines of each notebook.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description='Dump cell sources of Jupyter notebooks.',
            action=TOOL
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Dump cell sources of notebook.

        :param nb: notebook to dump
        :param nb_path: notebook's path
        :return: empty result
        """
        info_prefix = self._args.dump_info_prefix
        line_indent = ' ' * self._args.dump_source_line_indent
        cell_spacing = '\n' * self._args.dump_cell_spacing

        if self._args.dump_notebook_info:
            print('{} nbformat: {:1d}.{:1d}'
                  .format(info_prefix, nb.get('nbformat', 1), nb.get('nbformat_minor', 0)), end='')
            if 'kernelspec' in nb.metadata:
                print(' | kernel: {}'.format(nb.metadata.kernelspec.name), end='')
            if 'language_info' in nb.metadata:
                l_i = nb.metadata.language_info
                print(' | language: {} {}'.format(l_i.name, l_i.version), end='')
            print(' {}'.format(info_prefix))
            print('{} cells: {}'.format(info_prefix, len(nb.cells)), end='')
            other_metadata = nb_other_metadata(nb)
            if other_metadata:
                print(' | metadata: {}'.format(join_csv(other_metadata.keys())), end='')
            print(' {}'.format(info_prefix))

        spacing = ''

        for index, cell in enumerate(nb.cells):
            ct = cell.cell_type
            if ct in self._args.dump_cell_types:
                print(spacing, end='')
                spacing = cell_spacing
                if self._args.dump_cell_info:
                    print('{} cell {}: {}'.format(info_prefix, index, ct), end='')
                    metadata = set(cell.metadata.keys())
                    metadata.discard('tags')
                    if metadata:
                        print(' | metadata: {}'.format(','.join(metadata)), end='')
                    if 'tags' in cell.metadata:
                        tags = ','.join(cell.metadata.tags)
                        print(' | tags: {}'.format(tags), end='')
                    if (ct == 'markdown' or ct == 'raw') and 'attachments' in cell:
                        print(' | attachments: {}'.format(len(cell.attachments)), end='')
                    if ct == 'code' and cell.outputs:
                        print(' | outputs: {}'.format(len(cell.outputs)), end='')
                    print(' {}'.format(info_prefix))
                if self._args.dump_sources and cell.source:
                    for line_index, line in enumerate(cell.source.split('\n')):
                        if self._args.dump_source_line_numbers:
                            print('{}{:>2}| {}'.format(line_indent, line_index + 1, line), sep='')
                        else:
                            print('{}{}'.format(line_indent, line), sep='')

        return []  # no notebooks produced

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-g', '--notebook-info',
                           action=NegatableAction, default=SUPPRESS, dest='dump_notebook_info',
                           help='show global notebook information' +
                                ' (default: {})'.format(self._args.dump_notebook_info))
        group.add_argument('-c', '--cell-info',
                           action=NegatableAction, default=SUPPRESS, dest='dump_cell_info',
                           help='show cell information' +
                                ' (default: {})'.format(self._args.dump_cell_info))
        group.add_argument('-s', '--sources',
                           action=NegatableAction, default=SUPPRESS, dest='dump_sources',
                           help='show cell sources' +
                           ' (default: {})'.format(self._args.dump_sources))
        group.add_argument('-t', '--cell-types', default=SUPPRESS, type=split_csv, dest='dump_cell_types',
                           metavar='TYPES',
                           help='comma-separated list of cell types to dump' +
                                " (default: '{}')".format(join_csv(self._args.dump_cell_types)))
        group.add_argument('-p', '--prefix',
                           default=SUPPRESS, dest='dump_info_prefix', metavar='STRING',
                           help='pre/postfix for information lines' +
                                " (default: '{}')".format(self._args.dump_info_prefix))
        group.add_argument('-i', '--indent',
                           default=SUPPRESS, type=int, dest='dump_source_line_indent', metavar='INT',
                           help='indentation level for source lines' +
                                ' (default: {})'.format(self._args.dump_source_line_indent))
        group.add_argument('-l', '--line-numbers',
                           action=NegatableAction, default=SUPPRESS, dest='dump_source_line_numbers',
                           help='show source line numbers' +
                                ' (default: {})'.format(self._args.dump_source_line_numbers))
        group.add_argument('-e', '--cell-spacing',
                           default=SUPPRESS, type=int, dest='dump_cell_spacing', metavar='INT',
                           help='number of empty lines between cells' +
                                ' (default: {})'.format(self._args.dump_cell_spacing))

    def check_and_adjust_arguments(self) -> None:
        bad = set(self._args.dump_cell_types).difference(REQUIRED_CELL_FIELDS.keys())
        if bad:
            raise ValueError('Invalid cell types with -t: {}'.format(join_csv(bad)))

    def print_tool_args(self) -> None:
        print('  Dump notebook info: {}'.format(self._args.dump_notebook_info))
        print('  Dump cell info: {}'.format(self._args.dump_cell_info))
        print('  Dump cell sources: {}'.format(self._args.dump_sources))
        print('  Dump cell types: {}'.format(join_csv(self._args.dump_cell_types)))
        print('  Prefix for information lines: {}'.format(self._args.dump_info_prefix))
        print('  Source line indentation level: {}'.format(self._args.dump_source_line_indent))
        print('  Show source line numbers: {}'.format(self._args.dump_source_line_numbers))
        print('  Cell spacing lines: {}'.format(self._args.dump_cell_spacing))


def main(cli_args: List[str]=None):
    return DumpTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Functions for printing

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from typing import Any, Dict, Set
from argparse import Namespace

DEFAULT_WIDTH = 10


def print_set(s: Set[str], header: str=None, width: int=DEFAULT_WIDTH) -> None:
    """Print set s with section header.

    :param s: set to print
    :param header: header of the table
    :param width: width of the left column
    """
    if s:
        if header:
            print('{}:'.format(header))
        for element in sorted(s):
            print('  {:>{}} {}'.format('', width, element))


def print_dict(d: Dict[str, Any], header: str=None, width: int=DEFAULT_WIDTH) -> None:
    """Print dictionary d with section header.

    :param d: dictionary to print
    :param header: header of the table
    :param width: width of the left column
    """
    if d:
        if header:
            print('{}:'.format(header))
        for key in sorted(d):
            left = str(d[key])
            print('  {:>{}} {}'.format(left, width, key))


def print_namespace(ns: Namespace, header: str=None) -> None:
    """Print namespace with header.

    :param ns: namespace to print
    :param header: header of the table
    """
    if ns:
        if header:
            print('{}:'.format(header))
        keys = sorted(vars(ns))
        if keys:
            for var in keys:
                value = getattr(ns, var)
                print('  {} ({}): {!r}'.format(var, type(value).__name__, value))
        else:
            print('  (empty)')



"""
Tool nbvalidate

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import sys
from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup
from typing import List
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.validating import validate_nb

TOOL = 'validate'


class ValidateTool(Tool):
    """Validate each notebook, and report the validation results.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Validate Jupyter notebooks."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Validate notebook nb.

        :param nb: notebook to validate
        :param nb_path: path to ``nb``
        :return: empty result
        """
        valid = validate_nb(nb, getattr(self._args, 'assert'))

        if not valid or self._args.verbose:
            print('{} notebook structure is {}'.format(nb_path.name, 'valid' if valid else 'INVALID'))
        self._aggregate['outputs'][-1]['valid'] = valid

        return []

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        pass

    def print_tool_args(self) -> None:
        print('  Assert mode: {}'.format(getattr(self._args, 'assert')))


def main(cli_args: List[str]=None):
    return ValidateTool().main(cli_args)


if __name__ == "__main__":
    sys.exit(main())



"""
Functions for running

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import sys
from argparse import Namespace
from nbformat import NotebookNode
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors.execute import CellExecutionError
from .cleaning import clean_code_output, clean_code_metadata, truncate_output_streams

TEST = False


def run_nb(nb: NotebookNode, args: Namespace) -> None:
    """Run notebook.

    .. note:: **Modifies**: ``nb``

    :param nb: notebook to run
    :param args: arguments (options)
    """
    # clean up before execution
    if args.clean_before:
        clean_code_output(nb)

    if args.append_cell:
        nb.cells.append(nbformat.v4.new_code_cell(args.appended_cell))

    # run notebook
    ep = ExecutePreprocessor(
        timeout=args.timeout,
        kernel_name=args.kernel_name,
        allow_errors=args.allow_errors,
        interrupt_on_timeout=args.interrupt_on_timeout
    )
    try:
        resources = {'metadata': {'path': args.run_path}}  # set working directory
        _ = ep.preprocess(nb, resources)  # nb is executed in-place, locally
    except CellExecutionError as e:  # only possible if not args.allow_errors
        if getattr(args, 'assert'):  # args.assert gives syntax error
            raise
        else:
            print('{}: {}'.format(type(e).__name__, e), file=sys.stderr)
    finally:
        # clean up after execution
        if args.clean_after:
            clean_code_metadata(nb, args)

        if args.streams_head >= 0:
            truncate_output_streams(nb, args)



"""
Main entry point

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

# Command line argument handling inspired by nbdime.
# Config file mechanism: see config.py

import sys
from ._version import __version__, package_name

COMMANDS = ("config", "validate", "head", "dump", "stats", "view", "cat", "clean", "run", "split", "punch",)

HELP_MESSAGE_VERBOSE = ("Usage: nbtb [OPTIONS]\n\n"
                        "OPTIONS: -h, --version, COMMAND{%s}\n\n"
                        "Examples: nbtb --version\n"
                        "          nbtb config run\n"
                        "          nbtb stats -h\n"
                        "          nbtb clean -o \n\n"
                        "Documentation: https://nbtoolbelt.readthedocs.io"
                        % ", ".join(COMMANDS)
                        )


def main_dispatch(args=None) -> int:
    """Main entry point.

    :param args: command-line arguments (list of strings)
    :return: exit code
    """
    if args is None:
        args = sys.argv[1:]
    if len(args) < 1:
        sys.exit("Option missing.\n\n{}".format(HELP_MESSAGE_VERBOSE))

    cmd = args[0]
    args = args[1:]

    if cmd == 'config':
        from .config import load_config
        if len(args) == 0:
            tool = package_name
        elif len(args) == 1:
            tool = 'nb' + args[0]
        else:
            sys.exit("Too many arguments.\n\n{}".format(HELP_MESSAGE_VERBOSE))
        _ = load_config(tool, verbose=True)
        sys.exit(0)

    # TODO there should be a better way than the following way of dispatching
    elif cmd == "base":  # for testing of the Tool base class
        from .toolbaseapp import main
    elif cmd == "head":
        from .nbheadapp import main
    elif cmd == "dump":
        from .nbdumpapp import main
    elif cmd == "validate":
        from .nbvalidateapp import main
    elif cmd == "view":
        from nbtoolbelt.nbviewapp import main
    elif cmd == "stats":
        from nbtoolbelt.nbstatsapp import main
    elif cmd == "cat":
        from nbtoolbelt.nbcatapp import main
    elif cmd == "clean":
        from nbtoolbelt.nbcleanapp import main
    elif cmd == "run":
        from nbtoolbelt.nbrunapp import main
    elif cmd == "split":
        from nbtoolbelt.nbsplitapp import main
    elif cmd == "punch":
        from nbtoolbelt.nbpunchapp import main
    else:
        if cmd == '--version':
            sys.exit(__version__)
        if cmd == '-h' or cmd == '--help':
            sys.exit(HELP_MESSAGE_VERBOSE)
        else:
            sys.exit("Unrecognized command '%s'\n\n%s" %
                     (cmd, HELP_MESSAGE_VERBOSE))

    # use the imported main for selected command
    exit_code = main(args)
    if exit_code:
        print("Tool '{}' aborted.".format(cmd), file=sys.stderr)

    return exit_code


if __name__ == "__main__":
    # This is triggered by "python -m nbtoolbelt <args>"
    sys.exit(main_dispatch())



"""
Main for nbcat

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import Namespace
from typing import List
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.counting import nb_cell_stats, extract_aggregate
from nbtoolbelt.printing import print_dict

TOOL = 'cat'


class CatTool(Tool):
    """Catenate notebooks.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Catenate Jupyter notebooks into one notebook."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Collect cells of notebook nb.

        :param nb: notebook to catenate
        :param nb_path: path to ``nb``
        :return: catenation when called on last notebook
        """
        args = self._args

        if 'result' not in self._aggregate:
            self._aggregate['result'] = nb  # catenation of all notebooks so far
            if args.inplace:
                result_path = nb_path
            else:
                result_path = nb_path.with_name(nb_path.stem + args.cat_result_name + nb_path.suffix)
            self._aggregate['result_path'] = result_path
        else:
            self._aggregate['result'].cells.extend(nb.cells)

        # collect some statistics
        namespace = Namespace()
        setattr(namespace, 'cell_types', True)
        freq = nb_cell_stats(nb, namespace)['cell_types']

        # show statistics
        if not args.quiet:
            print_dict(freq, 'Counts')
        self._aggregate['outputs'][-1].update({'cell_types': freq})

        # on last file write result
        result = []
        if self._aggregate['outputs'][0]['file_count'] + 1 == len(self._args.notebooks):
            result.append((self._aggregate['result'], self._aggregate['result_path']))

        return result

    def process_collected_data(self):
        super().process_collected_data()
        totals = extract_aggregate(self._aggregate['outputs'][0]['cell_types'])
        print_dict(totals, 'Total counts')


def main(cli_args: List[str] = None):
    return CatTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Main for nbstats

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Summarize Jupyter notebooks on the command line.
"""

from nbtoolbelt.toolbaseapp import Tool
from textwrap import dedent
from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import List
from nbtoolbelt.arguments import NegatableAction
from nbformat import NotebookNode
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.counting import nb_metadata, nb_other_metadata, nb_extra_fields, nb_cell_stats, extract_aggregate
from nbtoolbelt.printing import print_dict

TOOL = 'stats'


class StatsTool(Tool):
    """Collect and show a statistical summary of each notebook.
    Also compute some statistics over all notebooks.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description=dedent("""\
                Summarize Jupyter notebooks on stdout with statistics.
                By default summarizes all notebook data.
                Limit to specific items by passing options.
                """),
            action='summarize'
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Collect and present statistics about notebook nb.

        :param nb: notebook to count
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        output = {}  # to accumulate JSON output

        # always print summary of required notebook metadata
        metadata = nb_metadata(nb)
        print_dict(metadata, 'Notebook metadata')
        output['notebook_metadata'] = metadata

        # print notebook-level optional metadata fields, if desired
        if args.metadata:
            other_metadata = nb_other_metadata(nb)
            print_dict(other_metadata, 'Other notebook metadata fields')
            output['notebook_other_metadata'] = other_metadata

        # print the notebook extra metadata, if desired
        if args.extra:
            extra_fields = nb_extra_fields(nb)
            print_dict(extra_fields, 'EXTRA NOTEBOOK FIELDS')
            output['notebook_extra_fields'] = extra_fields

        cell_stats = nb_cell_stats(nb, args)

        # print desired cell statistics
        keys = []
        if args.cell_types:
            print_dict(cell_stats['cell_types'], "Cell types")
            keys.append('cell_types')
        if args.sources:
            print_dict(cell_stats['sources'], "Cell sources")
            keys.append('sources')
        if args.metadata or args.tags:
            print_dict(cell_stats['cell_metadata'], "Cell metadata fields")
            keys.append('cell_metadata')
        if args.attachments:
            print_dict(cell_stats['cell_attachments'], "Cell attachments")
            keys.append('cell_attachments')
        if args.outputs or args.streams or args.errors:
            print_dict(cell_stats['code_outputs'], "Code cell outputs")
            keys.append('code_outputs')
        if args.execution:
            print_dict(cell_stats['code_execution'], 'Code cell execution')
            keys.append('code_execution')
        if args.extra:
            print_dict(cell_stats['cell_extra'], "EXTRA CELL FIELDS")
            keys.append('cell_extra')
        output.update({key: value for key, value in cell_stats.items() if key in keys})

        self._aggregate['outputs'][-1].update(output)

        return []

    def process_collected_data(self):
        super().process_collected_data()
        # print some aggregation results
        args = self._args

        if len(args.notebooks) <= 1:
            return

        # print header
        print('\nTotals\n======\n')

        aggregate = self._aggregate['outputs'][0]

        # print notebook-level optional metadata fields, if desired
        if args.metadata:
            other_metadata = extract_aggregate(aggregate.get('notebook_other_metadata', {}))
            print_dict(other_metadata, 'Other notebook metadata fields')

        # print the notebook extra metadata, if desired
        if args.extra:
            extra_fields = extract_aggregate(aggregate.get('notebook_extra_fields', {}))
            print_dict(extra_fields, 'EXTRA NOTEBOOK FIELDS')

        cell_stats = aggregate

        # print desired cell statistics
        if args.cell_types:
            cell_types = extract_aggregate(cell_stats.get('cell_types', {}))
            print_dict(cell_types, "Cell types")
        if args.sources:
            cell_sources = extract_aggregate(cell_stats.get('sources', {}))
            print_dict(cell_sources, "Cell sources")
        if args.metadata or args.tags:
            cell_metadata = extract_aggregate(cell_stats.get('cell_metadata', {}))
            print_dict(cell_metadata, "Cell metadata fields")
        if args.attachments:
            cell_attachments = extract_aggregate(cell_stats.get('cell_attachments', {}))
            print_dict(cell_attachments, "Cell attachments")
        if args.outputs or args.streams or args.errors:
            code_outputs = extract_aggregate(cell_stats.get('code_outputs', {}))
            print_dict(code_outputs, "Code cell outputs")
        if args.execution:
            # delete 'maximum In[#]' from collected outputs, because its statistics don't make much sense
            # TODO improve the following deletion; also: it mentions three attributes twice
            if 'code_execution' in cell_stats and 'maximum In[#]' in cell_stats['code_execution']:
                del cell_stats['code_execution']['maximum In[#]']
            code_execution = extract_aggregate(cell_stats.get('code_execution', {}))
            print_dict(code_execution, 'Code cell execution')
        if args.extra:
            cell_extra = extract_aggregate(cell_stats.get('cell_extra', {}))
            print_dict(cell_extra, "EXTRA CELL FIELDS")

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('--all',
                           action=NegatableAction, dest='all_stats', default=SUPPRESS,
                           help='show all statistics' +
                           ' (default: {})'.format(self._args.all_stats)),
        group.add_argument('-c', '--cell-types',
                           action=NegatableAction, default=SUPPRESS,
                           help='count cell types' +
                           ' (default: {})'.format(self._args.cell_types))
        group.add_argument('-s', '--sources',
                           action=NegatableAction, default=SUPPRESS,
                           help='statistics for cell sources' +
                           ' (default: {})'.format(self._args.sources))
        group.add_argument('-m', '--metadata',
                           action=NegatableAction, default=SUPPRESS,
                           help='show notebook metadata and count cell metadata' +
                           ' (default: {})'.format(self._args.metadata))
        group.add_argument('-t', '--tags',
                           action=NegatableAction, default=SUPPRESS,
                           help='count individual cell tags' +
                           ' (default: {})'.format(self._args.tags))
        group.add_argument('-a', '--attachments',
                           action=NegatableAction, default=SUPPRESS,
                           help='count cell attachment MIME types' +
                           ' (default: {})'.format(self._args.attachments))
        group.add_argument('-o', '--outputs',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell outputs' +
                           ' (default: {})'.format(self._args.outputs))
        group.add_argument('--streams',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell output stream names' +
                           ' (default: {})'.format(self._args.streams))
        group.add_argument('-e', '--errors',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell error names' +
                           ' (default: {})'.format(self._args.errors))
        group.add_argument('-x', '--execution',
                           action=NegatableAction, default=SUPPRESS,
                           help='statistics for code execution' +
                           ' (default: {})'.format(self._args.execution))
        group.add_argument('--extra',
                           action=NegatableAction, default=SUPPRESS,
                           help='report extra fields outside metadata' +
                           ' (default: {})'.format(self._args.extra))

    def check_and_adjust_arguments(self):
        args = self._args

        # if --all, summarize everything
        if args.all_stats:
            args.cell_types = True
            args.sources = True
            args.metadata = True
            args.tags = True
            args.attachments = True
            args.outputs = True
            args.streams = True
            args.errors = True
            args.execution = True
            args.extra = True


def main(cli_args: List[str]=None):
    return StatsTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Functions for cleaning

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from typing import Dict
from argparse import Namespace
from nbformat import NotebookNode
from collections import defaultdict
from .counting import count_source

TEST = False

# TODO split cleaning into smaller functions


def clean_nb(nb: NotebookNode, args: Namespace) -> Dict:
    """Clean fields and optionally output from notebook.

    If ``args`` is not ``None``, then the following arguments are used:

    * clean_notebook_metadata_fields (collection)
    * clean_cell_metadata_fields (collection)
    * clean_tags (collection)
    * clean_empty_cells (boolean)
    * clean_outputs (boolean)

    .. note:: **Modifies**: ``nb``

    :param nb: notebook to clean
    :param args: namespace with arguments (options)
    :return: dictionary with statistics for removed elements
    """
    freq = defaultdict(int)  # number of cleanings

    # delete notebook (global) metadata fields
    if args.clean_notebook_metadata_fields:
        for field in args.clean_notebook_metadata_fields:
            if field in nb.metadata:
                del nb.metadata[field]
                freq['global ' + field] += 1

    # delete empty cells, if desired
    n = len(nb.cells)
    if args.clean_empty_cells:
        nb.cells = [cell for cell in nb.cells if count_source(cell.source)[0]]
    if n > len(nb.cells):
        freq['empty cells'] = n - len(nb.cells)

    # traverse all cells, and delete fields
    for index, cell in enumerate(nb.cells):  # index is useful for debugging
        ct = cell.cell_type
        if TEST:
            print(ct, 'cell', index)

        # delete cell metadata fields
        if args.clean_cell_metadata_fields:
            for field in args.clean_cell_metadata_fields:
                if field in cell.metadata:
                    del cell.metadata[field]
                    freq['cell ' + field] += 1

        # delete cell tags
        if 'tags' in cell.metadata and args.clean_tags:
            removed_tags = {tag for tag in cell.metadata.tags if tag in args.clean_tags}
            for tag in removed_tags:
                freq['tag ' + tag] += 1
            clean_tags = [tag for tag in cell.metadata.tags if tag not in args.clean_tags]
            if clean_tags:
                cell.metadata.tags = clean_tags
            else:
                del cell.metadata['tags']

        # clean outputs of code cells, if requested
        if ct == 'code' and args.clean_outputs:
            if cell.outputs:
                cell.outputs = []
                freq['outputs'] += 1
            cell.execution_count = None

    return freq


def clean_code_output(nb: NotebookNode) -> None:
    """Clean all output cells and execution counts.

    .. note:: **Modifies**: ``nb``

    :param nb: notebook to clean
    """
    for cell in nb.cells:
        if 'outputs' in cell:
            cell.outputs = []
        if 'execution_count' in cell:
            cell.execution_count = None


def clean_code_metadata(nb: NotebookNode, args: Namespace) -> None:
    """Clean metadata of code cells, by
    * removing collapsed and scrolled flags on code cells
    * removing ExecuteTime data on code cells

    .. note::

        **Modifies**: ``nb``

    :param nb: notebook to clean
    :param args: namespace with arguments (options)
    """
    for cell in nb.cells:
        if cell.cell_type == 'code':
            md = cell.metadata
            for attr in args.clean_after_metadata:
                if attr in md:
                    del md[attr]


def truncate_output_streams(nb: NotebookNode, args: Namespace) -> None:
    """Limit the output cells to at most `args.limit_output` lines

    :param nb: notebook to limit
    :param args: namespace with arguments (options)
    """
    for cell in filter(lambda c: c.cell_type == 'code', nb.cells):
        head_per_stream = defaultdict(lambda: args.streams_head)

        new_outputs = []
        for output in cell.outputs:
            if output.output_type == 'stream':
                lines_remaining = head_per_stream[output.name]

                if lines_remaining > 0:
                    text = output.text

                    breakpoint = 0
                    while lines_remaining > 0:
                        next_newline = text.find("\n", breakpoint)
                        if next_newline < 0:
                            break

                        breakpoint = next_newline + 1
                        lines_remaining -= 1

                    if lines_remaining == 0:
                        output.text = text[:breakpoint] + "\n" + args.streams_truncate_message + "\n"

                    head_per_stream[output.name] = lines_remaining
                    new_outputs.append(output)

            else:
                new_outputs.append(output)

        cell.outputs = new_outputs



"""
Version, easily accessible in other files

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

# TODO: should each tool have its own version?

version_info = (2018, 1, 'dev0')
__version__ = ".".join(map(str, version_info))

package_name = 'nbtoolbelt'



"""
Functions for notebook reading and writing

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from pathlib import Path
from argparse import Namespace
from nbformat import NotebookNode
from typing import List, Union
import nbformat
import sys


def read_nb(nb_path: Path, args: Namespace) -> Union[None, NotebookNode]:
    """Read notebook from given path, and return it.
    Uses ``args.debug``: in debug mode, a read error results in an exception, else it returns ``None``.

    :param nb_path: path to read from
    :param args: to check debug mode
    :return: notebook read from ``nb_path`` or None if reading failed and not ``args.debug``
    """
    try:
        nb = nbformat.read(nb_path.open(encoding='utf-8'), as_version=4)
    except Exception as e:
        ename = type(e).__name__
        print('Reading of "{}" failed ({}):\n  {}'.format(nb_path.name, ename, e), file=sys.stderr)
        if args.debug:
            print('  Tried to read notebook from file:', nb_path.resolve(), file=sys.stderr)
            raise
        else:
            return None

    return nb


def write_nb(nb: NotebookNode, nb_path: Path, args: Namespace) -> Union[None, bool]:
    """Write given notebook to given path.
    Uses ``args.debug``: in debug mode, a write error results in an exception, else it returns ``None``.

    :param nb: notebook to be written
    :param nb_path: path to write to
    :param args: to check debug mode
    :return:
    """
    try:
        nbformat.write(nb, nb_path.open('w', encoding='utf-8'))
    except Exception as e:
        ename = type(e).__name__
        print('Writing of "{}" failed ({}):\n  {}.'.format(nb_path.name, ename, e), file=sys.stderr)
        if args.debug:
            print('  Tried to write notebook to file:', nb_path.resolve(), file=sys.stderr)
            raise
        else:
            return

    return True


def cell_lines(cell: NotebookNode) -> List[str]:
    """Return list of source lines for given cell.

    :param cell: cell whose source lines to return as list
    :return: list of source lines in cell
    """
    if cell.source:
        lines = cell.source.split('\n')
    else:
        lines = []

    return lines



"""
Tool nbview

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup, SUPPRESS
from typing import List
from nbtoolbelt.arguments import NegatableAction
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.rendering import render_nb
import webbrowser
import os

TOOL = 'view'


class ViewTool(Tool):
    """Show notebooks in the browser, with embedded MathJax and attachments rendered.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description='View Jupyter notebooks your browser.',
            action=TOOL
        )
        self._aggregate['written_html'] = []  # list of opened html files

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """View notebook.
        That is, convert it to html, and open it in the default browswer.

        .. note:: **Side-effect**: Creates html file

        :param nb: notebook to show head of
        :param nb_path: notebook's path
        :return: empty
        """
        # determine destination path of html file
        name_insert = '' if self._args.inplace else self._args.view_result_name
        html_path = nb_path.with_name(nb_path.stem + name_insert + '.html')

        # convert to html
        html = render_nb(nb, self._args)

        if self._args.write_files:
            # write html to file
            try:
                with html_path.open('w', encoding='utf-8') as fp:
                    fp.write(html)
            except Exception as e:
                ename = type(e).__name__
                print('Writing of "{}" failed ({}): {}.'.format(html_path.name, ename, e), file=sys.stderr)
                if self._args.debug:
                    print('  Tried to write html file:', html_path.resolve(), file=sys.stderr)
                raise

            # register in self._aggregate, for deletion (if so desired)
            if self._args.wait_delete:
                self._aggregate['written_html'].append(html_path)
        else:
            print('No file written')  # TODO is this really necessary?

        # open in browser
        if self._args.browser:
            webbrowser.open('file://' + Path(html_path).resolve().as_posix(), new=2)

        return []  # no notebooks produced

    def process_collected_data(self):
        if self._aggregate['written_html']:
            answer = input('Delete created html files ([y]/n)? ')
            deleting = not answer or answer.lower()[0] == 'y'
            for file in self._aggregate['written_html']:
                if deleting:
                    if not self._args.quiet:
                        print('Deleting:', file)
                    os.remove(file)
                else:
                    if not self._args.quiet:
                        print('Keeping:', file)

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-t', '--template-file', default=SUPPRESS,
                           metavar='FILE',
                           help='template file' +
                                ' (default: {})'.format(self._args.template_file))
        group.add_argument('-b', '--browser', action=NegatableAction, default=SUPPRESS,
                           help='open html in browser' +
                                ' (default: {})'.format(self._args.browser))
        group.add_argument('-w', '--wait_delete', action=NegatableAction, default=SUPPRESS,
                           help='interactively wait for user confirmation to delete html files' +
                                ' (default: {})'.format(self._args.wait_delete))

    def print_tool_args(self) -> None:
        print('  Template file: {!r}'.format(self._args.template_file))
        print('  Open in web browser: {!r}'.format(self._args.browser))
        print('  Wait for user to confirm deletion of html: {!r}'.format(self._args.wait_delete))

    def check_and_adjust_arguments(self):
        args = self._args
        if not args.write_files:
            args.browser = False
            args.wait_delete = False


def main(cli_args: List[str]=None):
    return ViewTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
NbConvert Preprocessor to inline attachments

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Parts of this code are reworked from the rendernb checklet for Momotor <momotor.org>.
"""

from nbconvert.preprocessors import Preprocessor
from nbformat import NotebookNode
from typing import Any, Dict, Tuple
import base64
from urllib.parse import quote


class InlineAttachmentsPreprocessor(Preprocessor):
    """To embed cell attachments inside the cell's source.
    """

    ACCEPTABLE_TYPES = (
        'image/png',
        'image/jpeg',
        'image/gif',
        'image/webp',
        'image/bmp',
    )

    def preprocess_cell(self, cell: NotebookNode, resources: Dict[str, Any], index: int) -> \
            Tuple[NotebookNode, Dict[str, Any]]:
        """Take a Jupyter Notebook cell and inline all attachments as data.

        .. note:: **Modifies**: ``cell``

        :param cell: cell for which to inline attachments
        :param resources: global resources
        :param index: cell index in notebook
        :return: modified cell, updated resources
        """
        if 'attachments' in cell:
            remaining_attachments = {}

            for name, data in cell.attachments.items():
                data_type, data_content = None, None
                # data_content = next((data[data_type] for data_type in data if data_type in ACCEPTABLE_TYPES), None)
                for data_type in data:
                    if data_type in self.ACCEPTABLE_TYPES:
                        data_content = data[data_type]
                        break

                if data_content:
                    # Re-encode base64 to strip whitespace
                    try:
                        decoded_content = base64.b64decode(data_content)
                    except TypeError:
                        pass
                    else:
                        recoded_content = base64.b64encode(decoded_content)

                    old_uri = 'attachment:' + quote(name)
                    new_uri = 'data:{};base64,{}'.format(quote(data_type), quote(recoded_content))

                    cell.source = cell.source.replace(old_uri, new_uri)

                else:
                    remaining_attachments[name] = data

            cell['attachments'] = remaining_attachments

        return cell, resources



"""
Common processing definitions

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from typing import Sequence, Tuple
from nbformat import NotebookNode
from pathlib import Path

# type alias for processing result
ProcessingResultType = Sequence[Tuple[NotebookNode, Path]]



"""
Tool nbpunch

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup, SUPPRESS
from typing import List
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.punching import punch_via_tags, PunchedNotebookProcessor, ChadsNotebookProcessor
from nbtoolbelt.printing import print_dict
from nbtoolbelt.arguments import split_csv, join_csv, NegatableAction, LoadSourceFile
from nbtoolbelt.counting import CELL_TYPES

TOOL = 'punch'


class PunchTool(Tool):
    """Punch notebooks, optionally filling holes with material coming from other notebooks,
    writing the results to new notebooks.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description = """Punch holes in Jupyter notebooks.""",
            action = TOOL
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Punch notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to punch
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args
        freq = {}

        if args.tags:
            freq, nb_punched, nb_chads = punch_via_tags(nb, args)

        # set up destination paths and result
        result = []

        if args.punched:
            if not args.tags:
                freq_punched, nb_punched = PunchedNotebookProcessor(nb, args, nb_path.name).punch_nb()
                freq.update(freq_punched)
            nb_punched_path = nb_path.with_name(nb_path.stem + args.punch_punched_result_name + nb_path.suffix)
            result.append((nb_punched, nb_punched_path))

        if args.chads:
            nb_chads_path = nb_path.with_name(nb_path.stem + args.punch_chads_result_name + nb_path.suffix)
            if not args.tags:
                freq_chads, nb_chads = ChadsNotebookProcessor(nb, args, nb_chads_path.name).punch_nb()
                freq.update(freq_chads)
            result.append((nb_chads, nb_chads_path))

        # print statistics
        if not args.quiet:
            print_dict(freq, 'Punch cell statistics')

        return result

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        args = self._args

        group.add_argument('-t', '--tags',
                           dest='tags', type=split_csv, default=SUPPRESS,
                           metavar='TAGS',
                           help='comma-separated list of tags that trigger removal of source from cells;' +
                                " e.g. YourTurn (default: '{}')".format(join_csv(args.tags)))
        group.add_argument('-p', '--punched',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to write punched notebook' +
                           ' (default: {})'.format(args.punched))
        group.add_argument('-c', '--chads',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to write chads notebook' +
                           ' (default: {})'.format(args.chads))
        group.add_argument('-m', '--keep-marker-lines',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to keep marker lines' +
                           ' (default: {})'.format(args.keep_marker_lines))
        group.add_argument('-f', '--fill',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to fill punched holes' +
                           ' (default: {})'.format(args.fill))
        group.add_argument('-s', '--source',
                           action=LoadSourceFile, dest='punch_source', default=SUPPRESS,
                           help='source notebook for filling holes; implies --fill' +
                           ' (default: "{}")'.format(args.punch_source))
        group.add_argument('-l', '--list',
                           action=NegatableAction, default=SUPPRESS,
                           help='list marker labels and descriptions' +
                           ' (default: {})'.format(args.list))
        group.add_argument('-e', '--allow-errors',
                           action=NegatableAction, default=SUPPRESS,
                           help='continue on errors' +
                           ' (default: {})'.format(args.allow_errors))

    def check_and_adjust_arguments(self):
        args = self._args

        if args.punch_source:
            # --source implies --fill
            args.fill = True

        if args.tags and args.punch_source:
            raise AssertionError('Cannot use --source with --tags.')

    def print_tool_args(self) -> None:
        args = self._args

        if args.tags:
            print('  Tags that trigger punch: {}'.format(args.tags))
        else:
            if args.debug:
                print('  Markers that trigger punch:')
                for cell_type in CELL_TYPES:
                    print('    {} regex: {}'.format(cell_type, repr(args.marker_regex[cell_type])))
                print('  Marker line parsing to extract transition, label, and description:')
                for cell_type in CELL_TYPES:
                    print('    {} regex: {}'.format(cell_type, repr(args.marker_line_parsing_regex[cell_type])))
                print('  Marker transitions:')
                for transition in 'begin', 'end':
                    print('    {}: {}'.format(transition, repr(args.marker_transitions[transition])))
        print('  Writing punched notebook: {}'.format(args.punched))
        print('  Writing chads notebook: {}'.format(args.chads))
        print('  Keeping marker lines: {}'.format(args.keep_marker_lines))
        print('  Listing labels and descriptions: {}'.format(args.list))
        if args.fill:
            if args.punch_source and not args.tags:
                print('  Fill holes from source notebook: {}'.format(args.punch_source))
            else:  # no source, or using tags
                print('  Fill holes as follows:')
                for cell_type in CELL_TYPES:
                    print('    {}: {}'.format(cell_type, repr(args.filling[cell_type])))
        else:
            print('  Do not fill punched holes')
        if args.allow_errors:
            action = 'Not stopping at errors'
        else:
            action = 'Stopping at first error'
        print('  {}'.format(action))


def main(cli_args: List[str]=None):
    return PunchTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Main for nbrun

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbtoolbelt.toolbaseapp import Tool
from nbformat import NotebookNode
from pathlib import Path
from argparse import _ArgumentGroup, SUPPRESS
from typing import List
from nbtoolbelt.arguments import NegatableAction, quote_help
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.running import run_nb
from nbtoolbelt.counting import nb_code_execution_stats
from nbtoolbelt.printing import print_dict

TOOL = 'run'


class RunTool(Tool):
    """Run each notebook, with option pre/post cleaning, and return results.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Run Jupyter notebooks, with optional pre/post-cleaning."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Run notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to run
        :param nb_path: path to ``nb``
        :return: resulting (notebook, notebook-path) pair
        """
        args = self._args

        run_nb(nb, args)

        # print statistics, unless in quiet mode
        stats = nb_code_execution_stats(nb)
        if not args.quiet:
            print_dict(stats, "Execution statistics")
        self._aggregate['outputs'][-1].update(stats)

        # set up destination path
        if args.inplace:
            nb_run_path = nb_path
        else:
            nb_run_path = nb_path.with_name(nb_path.stem + args.run_result_name + nb_path.suffix)

        if args.debug:  # TODO is this needed here (or should it happen in Tool.process_file?
            print('Destination of run:', nb_run_path)

        return [(nb, nb_run_path)]

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-e', '--allow-errors',
                           action=NegatableAction,
                           help='continue on errors' +
                                ' (default: {})'.format(self._args.allow_errors))
        group.add_argument('-b', '--clean-before',
                           action=NegatableAction,
                           help='clean code output before running' +
                                ' (default: {}'.format(self._args.clean_before))
        group.add_argument('-a', '--clean-after',
                           action=NegatableAction,
                           help='clean code metadata after running' +
                                ' (default: {})'.format(self._args.clean_after))
        group.add_argument('-k', '--kernel-name',
                           dest='kernel_name', default=SUPPRESS,
                           help='name of kernel to use, e.g. python3; '' for kernel specified in notebook;' +
                                " default: '{}')".format(self._args.kernel_name))
        group.add_argument('-p', '--run-path',
                           dest='run_path', default=SUPPRESS,
                           help='working directory for execution ;"" for current working directory' +
                                " (default: '{}')".format(self._args.run_path))
        group.add_argument('-t', '--timeout',
                           type=int, default=SUPPRESS,
                           help='timeout time in seconds per cell; -1 for unlimited' +
                                ' (default: {})'.format(self._args.timeout))
        group.add_argument('-i', '--interrupt-on-timeout',
                           action=NegatableAction, default=SUPPRESS,
                           help='interrupt execution on timeout' +
                                ' (default: {})'.format(self._args.interrupt_on_timeout))
        group.add_argument('-w', '--append-cell',
                           action=NegatableAction, default=SUPPRESS,
                           help='append special code cell' +
                                " (default: '{}') ".format(quote_help(self._args.appended_cell)) +
                                'before executing notebooks' +
                                ' (default: {})'.format(self._args.append_cell))
        group.add_argument('--streams-head',
                           type=int, default=SUPPRESS,
                           help='limit output streams to a maximum number of lines per cell; -1 for unlimited' +
                                ' (default: {})'.format(self._args.streams_head))

    def print_tool_args(self) -> None:
        if self._args.clean_before:
            action = 'Cleaning'
        else:
            action = 'Keeping'
        print('  {} code output before running'.format(action))

        print('  Timeout:', 'unlimited' if self._args.timeout < 0 else '{} s'.format(self._args.timeout))
        print('  Kernel name:', self._args.kernel_name if self._args.kernel_name else 'from notebook')
        print('  Working dir:', self._args.run_path if self._args.run_path else 'current directory')

        if self._args.allow_errors:
            action = 'Not stopping at errors'
        else:
            action = 'Stopping at first error'
        print('  {}'.format(action))

        if self._args.append_cell:
            action = 'Appending'
        else:
            action = 'Not appending'
        print('  {} code cell with %whos'.format(action))

        if self._args.clean_after:
            action = 'Cleaning {}'.format(self._args.clean_after_metadata)
        else:
            action = 'Keeping'
        print('  {} code metadata after running'.format(action))


def main(cli_args: List[str] = None):
    return RunTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Functions for validating notebooks

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from nbformat import NotebookNode
import nbformat


def validate_nb(nb: NotebookNode, assertive: bool) -> bool:
    """Validate notebook.

    :param nb: notebook to validate
    :param assertive: whether to raise an exception when invalid
    :return: whether ``nb`` is valid
    """
    try:
        nbformat.validate(nb)
        return True
    except nbformat.ValidationError:
        if assertive:
            raise
        else:
            return False



"""
Setup file for nbtoolbelt

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from setuptools import setup, find_packages
import sys
from pathlib import Path

here = Path(__file__).parent  # location of this file in the file system

version_ns = {}  # name space for version
with (here / 'src' / 'nbtoolbelt' / '_version.py').open(encoding="utf8") as f:
    exec(f.read(), {}, version_ns)

#-----------------------------------------------------------------------------
# Minimal Python version sanity check
#-----------------------------------------------------------------------------

v = sys.version_info
if v[0] < 3 or (v[0] >= 3 and v[:2] < (3, 5)):
    error = "ERROR: %s requires Python version 3.5 or above." % name
    print(error, file=sys.stderr)
    sys.exit(1)

#-----------------------------------------------------------------------------
# get on with it
#-----------------------------------------------------------------------------

# I really prefer Markdown to reStructuredText.  PyPi does not.  This allows me
# to have things how I'd like, but not throw complaints when people are trying
# to install the package and they don't have pypandoc or the README in the
# right place.
try:
   import pypandoc
   long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
   long_description = ''

setup(
    name = version_ns['package_name'],
    version = version_ns['__version__'],
    author = 'Tom Verhoeff',
    author_email = 't.verhoeff@tue.nl',
    description = "Tools to work with Jupyter notebooks",
    long_description=long_description,
    url = 'https://gitlab.tue.nl/jupyter-projects/nbtoolbelt',
    license = 'MIT License',
    platforms = 'Linux, Mac OS X, Windows',
    keywords = ['Interactive', 'Interpreter', 'Shell', ],
    packages = find_packages('src'),
    install_requires = [
        'nbformat',
        'nbconvert',
        'numpy',
        'pandas',
    ],
    python_requires = '>=3.5',
    tests_require = {
        'test': ['pytest', 'pytest-mock'],
    },
    package_dir = {'': 'src'},
    package_data = {'': [
        'data/nbtoolbelt.json'
    ]},
    entry_points = {
        'console_scripts' : [
            'nbtb = nbtoolbelt.__main__:main_dispatch',
        ]
    },
    # zip_safe = False,
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Framework :: Jupyter',
    ],
)



"""
Functions for command-line argument parsing

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import argparse
from typing import List, Iterable
from pathlib import Path
from .config import load_config_file
from .printing import print_namespace
from .notebook_io import read_nb
from .punching import SourceChadsProcessor

TEST = False


class TestAction(argparse.Action):
    """Action to test the action mechanism,
    by printing the parameter values received in constructor and call methods.
    """
    def __init__(self, option_strings, dest, **kwargs):
        """Invoked when argument is configured in argmument parser.
        Some/all parameters of ``add_argument`` are passed on via ``kwargs`` and
        are stored in ``self``.
        """
        super().__init__(option_strings, dest, **kwargs)
        print('TestAction.__init__(option_strings={}, dest={}, **kwargs)'.format(option_strings, dest))
        print('  kwargs={}'.format(kwargs))

    def __call__(self, parser, namespace, values, option_string, **kwargs) -> None:
        """Invoked when argument is encountered during parsing.

        .. note:: **Modifies**: ``ns``

        :param parser: parser at work
        :param namespace: namespace to inspect/modify
        :param values: value for triggering option (already processed)
        :param option_string: option that triggered the action
        """
        items = [values, option_string]
        print('TestAction.__call__(parser, namespace, values={}, option_string={})'.format(*items))
        print_namespace(namespace, 'namespace')
        print_namespace(self, 'self')


class LoadConfigFile(argparse.Action):
    """Action to take when parsing a configuration file option.
    """
    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None) -> None:
        """Load configuration file for tool mentioned in dest.

        .. note:: **Modifies**: ``namespace``

        :param parser: parser at work
        :param namespace: namespace to inspect/modify
        :param values: value for triggering option (already processed)
        :param option_string: option that triggered the action
        """
        load_config_file(self.const, namespace, values, True)
        if hasattr(namespace, self.dest):
            delattr(namespace, self.dest)  # clean up the namespace; attribute 'config' is not used


class ExtendAction(argparse.Action):
    """Extend current list for option with parsed list.

    .. note:: **Assumption**:
      Option is already present in namespace (from embedded config file read in ``__init__()``.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, ns, values, option_string=None):
        """Extend dest in ns with values.

        .. note:: **Modifies**: ``ns``
        """
        getattr(ns, self.dest).extend(values)


class NegatableAction(argparse.Action):
    """Add the supplied positive options and negative versions as well.
    Inspired by ``nbshow``.
    """

    def __init__(self, option_strings, dest, default=None, required=False, help=None):
        opts = []
        for opt in option_strings:
            if len(opt) == 2 and opt[0] == '-':
                if not opt[1].islower():
                    raise ValueError('Single character flags should be lower-case for NegatableAction')
                opts.append(opt)
                opts.append(opt.upper())
            elif opt[:2] == '--':
                opts.append(opt)
                opts.append('--no-' + opt[2:])
            else:
                ValueError('Could not turn option "%s" into an NegatableAction option.')

        # Put positives first, negatives last:
        opts = opts[0::2] + opts[1::2]

        super().__init__(
            opts, dest, nargs=0, const=None,
            default=default, required=required,
            help=help)

    def __call__(self, parser, ns, values, option_string=None):
        """Store bool derived from option_string into dest of ns.

        .. note:: **Modifies**: ``ns``
        """
        if len(option_string) == 2:
            setattr(ns, self.dest, option_string[1].islower())
        else:
            setattr(ns, self.dest, option_string[2:2 + len('no')] != 'no')


class LoadSourceFile(argparse.Action):
    """Action to take when parsing a source file option.
    """
    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None) -> None:
        """Load source file into namespace.source_chads.

        .. note:: **Modifies**: ``namespace``

        :param parser: parser at work
        :param namespace: namespace to inspect/modify
        :param values: value for triggering option (already processed)
        :param option_string: option that triggered the action
        """
        # read notebook whose file name is in values
        setattr(namespace, self.dest, values)
        nb = read_nb(Path(values), namespace)
        if nb is None:
            raise AssertionError('Could not read source chads')
        freq = SourceChadsProcessor(nb, namespace, values).punch_nb()
        if namespace.debug or not namespace.source_chads:
            print('{} chads loaded from source notebook "{}".'
                  .format(freq['holes'], values))
        if TEST:
            import pprint
            pprint.pprint(getattr(namespace, self.dest))


def split_csv(arg: str) -> List[str]:
    """Split comma-separated values into a list.

    :param arg: argument to split
    :return: list of separated valued
    """
    return arg.split(',')  # returning list, and not set, to allow reading from JSON config file


def join_csv(s: Iterable[str]) -> str:
    """Join strings in list to comma-separated values.
    Inverse of ``split_csv``.

    :param s: list to join
    :return: joined csv string
    """
    return ','.join(s)


def quote_help(s: str) -> str:
    """Double percent signs, and quote newlines.
    This is needed in help for argparse.
    """
    return s.replace('%', '%%').replace('\n','\\n')
